/** @file maskdila.c
 *  @brief Mask image dilation.
 *  @copyright (c) Turku PET Centre
 *  @author Vesa Oikonen
 */
/// @cond
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
/*****************************************************************************/
#include "libtpcmisc.h"
#include "libtpcimgio.h"
#include "libtpcimgp.h"
/*****************************************************************************/

/*****************************************************************************/
static char *info[] = {
  "Dilate mask image in ECAT 6.3 or 7.x, NIfTI-1, or Analyze 7.5 format.",
  "Notice that mask image file is overwritten; make sure that backup exists.",
  " ",
  "Usage: @P [Options] maskfile",
  " ",
  "Options:",
  " -struct=<<cube>|<rcube>|<star>>",
  "     Select the structuring element for dilation; either 3x3x3 cube (default),",
  "     rounded cube (3x3x3 cube without corners), or star (cube on its corner,",
  "     consisting of 7 voxels).",
  " -stdoptions", // List standard options like --help, -v, etc
  " ",
  "See also: maskeros, imgmask, masksize, imgthrs, imgmax, imgbox",
  " ",
  "Keywords: image, mask, dilation, threshold",
  0};
/*****************************************************************************/

/*****************************************************************************/
/* Turn on the globbing of the command line, since it is disabled by default in
   mingw-w64 (_dowildcard=0); in MinGW32 define _CRT_glob instead, if necessary;
   In Unix&Linux wildcard command line processing is enabled by default. */
/*
#undef _CRT_glob
#define _CRT_glob -1
*/
int _dowildcard = -1;
/*****************************************************************************/

/*****************************************************************************/
/**
 *  Main
 */
int main(int argc, char **argv)
{
  int     ai, help=0, version=0, verbose=1;
  int     ret, n;
  char    maskfile[FILENAME_MAX], imgfile[FILENAME_MAX];
  int     structuring_element=1; // 1=cube, 2=rcube, 3=star
  char   *cptr=NULL;
  unsigned int mn;


  /*
   *  Get arguments
   */
  if(argc==1) {tpcPrintUsage(argv[0], info, stderr); return(1);}
  maskfile[0]=imgfile[0]=(char)0;
  /* Options */
  for(ai=1; ai<argc; ai++) if(*argv[ai]=='-') { /* options */
    cptr=argv[ai]+1; if(*cptr=='-') cptr++; if(cptr==NULL) continue;
    if(tpcProcessStdOptions(argv[ai], &help, &version, &verbose)==0) continue;
    if(strncasecmp(cptr, "STRUCT=", 7)==0) {
      cptr+=7;
      if(strncasecmp(cptr, "CUBE", 1)==0) {structuring_element=1; continue;}
      if(strncasecmp(cptr, "RCUBE", 1)==0) {structuring_element=2; continue;}
      if(strncasecmp(cptr, "STAR", 1)==0) {structuring_element=3; continue;}
    }
    fprintf(stderr, "Error: invalid option '%s'.\n", argv[ai]); 
    return(1);
  } else break;
  
  /* Print help or version? */
  if(help==2) {tpcHtmlUsage(argv[0], info, ""); return(0);}
  if(help) {tpcPrintUsage(argv[0], info, stdout); return(0);}
  if(version) {tpcPrintBuild(argv[0], stdout); return(0);}

  /* Process other arguments, starting from the first non-option */
  if(ai<argc) {strlcpy(maskfile, argv[ai], FILENAME_MAX); ai++;}
  if(ai<argc) {strlcpy(imgfile, argv[ai], FILENAME_MAX); ai++;}
  if(ai<argc) {fprintf(stderr, "Error: too many arguments.\n"); return(1);}

  /* Did we get all the information that we need? */
  if(!maskfile[0]) {
    fprintf(stderr, "Error: missing command-line argument; use option --help\n");
    return(1);
  }


  /* In verbose mode print options */
  if(verbose>1) {
    printf("maskfile := %s\n", maskfile);
    if(imgfile[0]) printf("imgfile := %s\n", imgfile);
    printf("structuring_element := %d\n", structuring_element);
    fflush(stdout);
  }


  /*
   *  Read mask, and optional PET image
   */
  IMG mask, img; imgInit(&mask); imgInit(&img);

  if(verbose>0) {printf("reading %s\n", maskfile); fflush(stdout);}
  ret=imgRead(maskfile, &mask);
  if(ret) {
    fprintf(stderr, "Error: %s\n", mask.statmsg);
    if(verbose>1) printf("ret := %d\n", ret);
    return(2);
  }
  if(mask.dimt>1) {
    fprintf(stderr, "Error: mask cannot be dynamic image.\n");
    imgEmpty(&mask); return(2);
  }
  /* Check whether mask has any voxels to begin with */
  mn=imgMaskCount(&mask);
  if(mn==0) {
    fprintf(stderr, "Warning: initial mask contains no positive voxels.\n");
    imgEmpty(&mask);
    return(0);
  }
  if(mn==(unsigned int)(mask.dimz*mask.dimy*mask.dimx)) {
    fprintf(stderr, "Warning: initial mask contains all positive voxels.\n");
    imgEmpty(&mask);
    return(0);
  }
  if(verbose>1) {
    printf("initial_nr_of_positive_voxels := %ud\n", mn);
  }

  if(imgfile[0]) {
    if(verbose>0) {printf("reading %s\n", imgfile); fflush(stdout);}
    ret=imgRead(imgfile, &img);
    if(ret) {
      fprintf(stderr, "Error: %s\n", img.statmsg);
      if(verbose>1) printf("ret := %d\n", ret);
      imgEmpty(&mask); return(3);
    }
    /* Check that dimensions are compatible */
    if(img.dimx!=mask.dimx || img.dimy!=mask.dimy) {
      fprintf(stderr, "Error: different image dimensions.\n");
      if(verbose>0) {
        printf("image := %d x %d\n", img.dimx, img.dimy);
        printf("mask := %d x %d\n", mask.dimx, mask.dimy);
      }
      imgEmpty(&img); imgEmpty(&mask); return(4);
    }
    if(mask.dimz!=1 && img.dimz!=mask.dimz) {
      fprintf(stderr, "Error: different image dimensions.\n");
      if(verbose>0) {
        printf("image := %d x %d x %d\n", img.dimx, img.dimy, img.dimz);
        printf("mask := %d x %d x %d\n", mask.dimx, mask.dimy, mask.dimz);
      }
      imgEmpty(&img); imgEmpty(&mask); return(4);
    }
    if(mask.dimt!=1 && img.dimt!=mask.dimt) {
      fprintf(stderr, "Error: different image dimensions.\n");
      if(verbose>0) {
        printf("image := %d x %d x %d x %d\n", 
               img.dimx, img.dimy, img.dimz, img.dimt);
        printf("mask := %d x %d x %d x %d\n", 
               mask.dimx, mask.dimy, mask.dimz, mask.dimt);
      }
      imgEmpty(&img); imgEmpty(&mask); return(4);
    }
  }

  /*
   *  Make the structuring element (dimensions must be odd)
   */
  IMG selem; imgInit(&selem);
  ret=imgStructuringElement(&selem, structuring_element, verbose-1);
  if(ret!=0) {
    fprintf(stderr, "Error: cannot make structuring element.\n");
    imgEmpty(&img); imgEmpty(&mask); imgEmpty(&selem);
    return(5);
  }


  /*
   *  Dilate the mask
   */
  if(verbose>0) {printf("dilating\n"); fflush(stdout);}
  if(!imgfile[0])
    n=imgMaskDilate(&mask, &selem);
  else
    n=-1;
  if(n<0) {
    fprintf(stderr, "Error: cannot dilate the mask.\n");
    if(verbose>1) printf("ret := %d\n", -n);
    imgEmpty(&img); imgEmpty(&mask); imgEmpty(&selem);
    return(8);
  }
  if(n==0) {
    fprintf(stdout, "No voxels dilated.\n");
    imgEmpty(&img); imgEmpty(&mask); imgEmpty(&selem);
    return(0);
  }
  if(verbose>0) {
    fprintf(stdout, "%d mask voxel(s) dilated.\n", n);
    fflush(stdout);
  }
  /* Check the number of masked voxels  */
  mn=imgMaskCount(&mask);
  if(mn==0) {
    fprintf(stderr, "Warning: empty mask, not saved.\n");
    imgEmpty(&img); imgEmpty(&mask); imgEmpty(&selem);
    return(0);
  }
  if(verbose>1) {
    printf("nr_of_positive_voxels := %u\n", mn);
  }
  if(mn==(unsigned int)(mask.dimz*mask.dimy*mask.dimx)) {
    fprintf(stderr, "Warning: dilated mask contains all positive voxels.\n");
  }  
  imgEmpty(&img); imgEmpty(&selem);

  /*
   *  Save the modified mask
   */
  if(verbose>2) printf("writing mask\n");
  ret=imgWrite(maskfile, &mask);
  if(ret) {
    fprintf(stderr, "Error: %s\n", mask.statmsg);
    imgEmpty(&mask); return(11);
  }
  imgEmpty(&mask);
  if(verbose>0) printf("mask written.\n\n");

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/// @endcond
