/******************************************************************************

  Copyright (c) 2008,2009 by Turku PET Centre

  Library:     scanditronics.c
  Description: Processing Scanditronics (and GEMS) blood on-line sampler data. 

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
  Place, Suite 330, Boston, MA 02111-1307 USA.

  Turku PET Centre hereby disclaims all copyright interest in the program.
  Juhani Knuuti
  Director, Professor
  Turku PET Centre, Turku, Finland, http://www.turkupetcentre.fi/

  Modification history:
  2008-12-13 Vesa Oikonen
             First created.
  2008-12-14 Vesa Oikonen
             Added bdata_coinc_count() and bdata_write().
             bdata_read() checks for catenated files.
  2009-03-16 VO
             bdata_get_date() reads time from Scanditronics header if it exists
             there. 
             GEMS time is always read from first sample.  
             Added functions bdata_get_start_time() and bdata_correct_header().
             Uses new enums.
             Filename corrected to scanditronics.c.

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
/*****************************************************************************/
#include "scanditronics.h"
/*****************************************************************************/
/* Local functions */
int ch_mean_comp(const void *, const void *);
int sample_time_comp(const void *i, const void *j);
/*****************************************************************************/

/*****************************************************************************/
/** Initiate BData structure. This should be called once before use. */
void bdata_init(BData *data)
{
  memset(data, 0, sizeof(BData));
  data->sampler_id=BSAMPLER_UNKNOWN;
  data->format=BFORMAT_UNKNOWN;
  data->sample_nr=0;
  data->sample_space=0;
  data->header_nr=0;
  data->sample=NULL;
  data->start_time=(time_t)0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Free memory allocated for BData. All data is cleared. */
void bdata_empty(BData *data)
{
  data->sampler_id=BSAMPLER_UNKNOWN;
  data->format=BFORMAT_UNKNOWN;
  if(data->sample_space>0) free((char*)(data->sample));
  data->sample_space=0;
  data->sample_nr=0;
  data->header_nr=0;
  data->start_time=(time_t)0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Allocate memory for specified nr of samples */
int bdata_allocate(BData *data, int size)
{
  if(data==NULL || size<1) return 1;
  if(data->sample_space>=size) return 0;
  if(data->sample_space==0) {
    data->sample=(BSample*)calloc(size, sizeof(BSample));
  } else {
    data->sample=(BSample*)realloc(data->sample, size*sizeof(BSample));
  }
  if(data->sample==NULL) {
    data->sample_space=0;
    return 2;
  }
  data->sample_space=size;
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Read Scanditronics or GEMS blood online sampler data file.
\return Returns 0 when successful. 
 */
int bdata_read(
  /** Filename of blood data */
  char *filename,
  /** Pointer to initiated data struct where data will be written.
   *  Any previous contents are deleted. */
  BData *bdata,
  /** Verbose level; set to zero to not to print any comments */
  int verbose,
  /** Pointer to a string (allocated for at least 64 chars) where error message 
   *  or other execution status will be written; enter NULL, if not needed */
  char *status
) {
  char ch, *cptr, line[MAX_SAMPLER_LINE_LEN], temp[256];
  FILE *fp;
  int i, n, ret;
  double d[8];

  /* Check input */
  if(status!=NULL) strcpy(status, "invalid input");
  if(filename==NULL || strlen(filename)==0) return 1;
  if(bdata==NULL) return 1;
  if(verbose>0) {
    printf("bdata_read(%s, *bdata, %d, status)\n", filename, verbose);
    fflush(stdout);
  }

  /* Delete any previous data */
  bdata_empty(bdata);

  /* Open bloodfile */
  if(verbose>0) printf("reading file %s\n", filename);
  if((fp=fopen(filename, "r")) == NULL) {
    if(status!=NULL) sprintf(status, "cannot open file");
    return 2;
  }

  /* Count the nr of lines and allocate memory accordingly */
  i=0; do {if((ch=fgetc(fp))=='\n') i++;} while(ch != EOF);
  if(verbose>1) printf("lines := %d\n", i);
  /* Allocate memory for sampler data, plus some extra space */
  ret=bdata_allocate(bdata, i+2);
  if(ret!=0) {
    if(status!=NULL) sprintf(status, "cannot allocate memory");
    fclose(fp); return 3;
  }

  /* Try to determine the sampler id and format from filename */
  n=strlen(filename);
  if(n>4) {
    cptr=filename+n-4;
    if(strcasecmp(cptr, ".lis")==0) {
      bdata->sampler_id=BSAMPLER_SCANDITRONICS;
      bdata->format=BFORMAT_SCANDITRONICS;
    } else if(strcasecmp(cptr, ".bld")==0) {
      bdata->sampler_id=BSAMPLER_GEMS;
      bdata->format=BFORMAT_SCANDITRONICS;
    } else if(strcasecmp(cptr, ".alg")==0) {
      bdata->sampler_id=BSAMPLER_ALLOGG;
      bdata->format=BFORMAT_ALLOGG;
    }
  }
  if(verbose>1) {
    printf("filename_based_sampler_id := %d\n", bdata->sampler_id);
    printf("filename_based_format := %d\n", bdata->format);
  }

  /* Read file line-by-line */
  rewind(fp); bdata->sample_nr=0; bdata->header_nr=0;
  while(fgets(line, MAX_SAMPLER_LINE_LEN-1, fp) != NULL) {
    if(verbose>4) {fprintf(stdout, "%s", line); fflush(stdout);}
    n=strlen(line); if(n<2) continue;
    /* Get the first non-space character */
    cptr=line; while(isspace(*cptr)) cptr++;
    n=strlen(cptr); if(n<2) continue;
    /* get comment lines */
    if(*cptr=='#' || *cptr==';' || *cptr=='*' || strncmp(cptr, "//", 2)==0) {
      if(bdata->header_nr>=MAX_SAMPLER_HEADERS) {
        if(status!=NULL) sprintf(status, "too big header");
        fclose(fp); return 4;
      }
      strcpy(bdata->header[bdata->header_nr++], cptr);
      continue;
    }
    /* get blood sample lines */
    n=sscanf(cptr,
          "%lf %lf %lf %i %i %i %i %i %i %i",
          &bdata->sample[bdata->sample_nr].time_of_day,
          &bdata->sample[bdata->sample_nr].time_since_start,
          &bdata->sample[bdata->sample_nr].meas_interval,
          &bdata->sample[bdata->sample_nr].pair1_coincid,
          &bdata->sample[bdata->sample_nr].pair1_det1_cnts,
          &bdata->sample[bdata->sample_nr].pair1_det2_cnts,
          &bdata->sample[bdata->sample_nr].pair2_coincid,
          &bdata->sample[bdata->sample_nr].pair2_det1_cnts,
          &bdata->sample[bdata->sample_nr].pair2_det2_cnts,
          &bdata->sample[bdata->sample_nr].aux_cnts
        );
    if(n==10) {
      bdata->sample_nr++;
      if(bdata->sample_nr>=bdata->sample_space-1) {
        if(status!=NULL) sprintf(status, "too large data set");
        fclose(fp); bdata_empty(bdata); return 4;
      }
      continue;
    }
    /* Read GEMS information line as a comment */
    // us1321 2.050000 1.230000 1.400000 1.400000 1.400000 1.400000
    n=sscanf(line, "%s %lf %lf %lf %lf %lf %lf %lf",
           temp, &d[0], &d[1], &d[2], &d[3], &d[4], &d[5], &d[6]);
    if(n==7) {
      if(bdata->header_nr<MAX_SAMPLER_HEADERS) {
        strcpy(bdata->header[bdata->header_nr++], line);
        continue;
      } else {
        if(status!=NULL) sprintf(status, "too large data set");
        fclose(fp); bdata_empty(bdata); return 4;
      }
    }
    /* Line was not a comment nor a constant data line */
    if(verbose) {
      fprintf(stderr, "invalid line: %s", line); fflush(stderr);
    }
    /* Add it as a comment, if there is room for it */
    if(bdata->header_nr<MAX_SAMPLER_HEADERS) {
      strcpy(bdata->header[bdata->header_nr++], cptr);
      continue;
    }
    /* Otherwise consider it as an error */
    if(n!=10) {
      if(status!=NULL) sprintf(status, "invalid line");
      fclose(fp); bdata_empty(bdata); return 5;
    }
  } 
  fclose(fp);
  if(verbose) {
    fprintf(stdout, "%d header lines read\n", bdata->header_nr);
    fprintf(stdout, "%d sample lines read\n", bdata->sample_nr);
  }
  if(bdata->header_nr<1 || bdata->sample_nr<1) {
    if(status!=NULL) sprintf(status, "invalid file format");
    return 6;
  }
  /* Check that headers and data contain valid measurement date and time */
  if(bdata_get_start_time(bdata, verbose, status)!=0) {
    bdata_empty(bdata); return 7;
  }

  /* Check that time_since_start is increasing; if not, then user should
     check the file, it may be catenated */
  for(i=1, ret=0; i<bdata->sample_nr; i++) {
    if(bdata->sample[i].time_since_start>bdata->sample[i-1].time_since_start)
      continue;
    ret=1; break;
  }
  if(ret!=0) {
    if(status!=NULL) sprintf(status, "invalid file, revise manually");
    bdata_empty(bdata); return 8;
  }
  /* Calculate sample times and channel cps etc */
  ret=bdata_calc(bdata);
  if(ret) {
    if(status!=NULL) sprintf(status, "invalid count data");
    bdata_empty(bdata); return 9;
  }
  if(status!=NULL) sprintf(status, "ok");
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Calculate the measurement duration corrected activities, and
 *  the mean of coincidence channels.
\return Returns 0 if successful.
 */
int bdata_calc(BData *bdata)
{
  int i;

  if(bdata==NULL) return 1;
  for(i=0; i<bdata->sample_nr; i++) {
    bdata->sample[i].t=
      bdata->sample[i].time_since_start+0.5*bdata->sample[i].meas_interval;
    bdata->sample[i].ch1=(double)bdata->sample[i].pair1_coincid;
    bdata->sample[i].ch2=(double)bdata->sample[i].pair2_coincid;
    if(bdata->sample[i].meas_interval>0) {
      bdata->sample[i].ch1/=(double)bdata->sample[i].meas_interval;
      bdata->sample[i].ch2/=(double)bdata->sample[i].meas_interval;
    }
    bdata->sample[i].ch_mean=0.5*(bdata->sample[i].ch1+bdata->sample[i].ch2);
  }
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Examines whether detector channel ratio exceeds a user-specified limit.
 *  Ratio is calculated from ten highest overall samples.
\return Returns <0 in case of an error, 0 if channels are similar enough,
        or 1 in case the ratio exceeds the limit.
*/
int bdata_examine(
  /** Sampler data to be examined */
  BData *bdata,
  /** Lower-to-higher channel limit ratio, e.g. 0.7 */
  double limit,
  /** Channel ratio will be written here, NULL if not required */
  double *r,
  /** Channel 1 mean from ten highest overall samples, NULL if not required */
  double *ch1,
  /** Channel 2 mean from ten highest overall samples, NULL if not required */
  double *ch2
) {
  int i;
  double ch1_mean, ch2_mean, ch_ratio;

  if(bdata==NULL || limit<0.0 || limit>1.0) return -1;
  if(bdata->sample_nr<10) return -2;

  /* Sort the data by decreasing channel mean;
     remember to sort by time before returning */
  qsort(bdata->sample, bdata->sample_nr, sizeof(BSample), ch_mean_comp);

  /* Calculate the mean of ten highest activities separately for both channels */
  ch1_mean=ch2_mean=0.0;
  for(i=0; i<10; i++) {
    ch1_mean+=bdata->sample[i].ch1;
    ch2_mean+=bdata->sample[i].ch2;
  }
  ch1_mean/=10.0; ch2_mean/=10.0;
  if(ch1!=NULL) *ch1=ch1_mean;
  if(ch2!=NULL) *ch2=ch2_mean;

  /* Calculate the channel ratio */
  ch_ratio=1.0;
  if(ch1_mean > ch2_mean) {
    if(fabs(ch1_mean)>1.0E-10) ch_ratio = ch2_mean / ch1_mean;
    else ch_ratio = 0.0;
  } else {
    if(fabs(ch2_mean)>1.0E-10) ch_ratio = ch1_mean / ch2_mean;
    else ch_ratio = 0.0;
  }
  if(r!=NULL) *r=ch_ratio;

  /* Sort the data by increasing sample time */
  qsort(bdata->sample, bdata->sample_nr, sizeof(BSample), sample_time_comp);

  /* Check the ratio against the limit */
  if(ch_ratio<limit) return 1;
  return 0;
}

/* Compare the channel means */
int ch_mean_comp(const void *i, const void *j)
{
  BSample *d1, *d2;
  d1=(BSample *)i; d2=(BSample *)j;

  return(d2->ch_mean - d1->ch_mean);
}

/* Compare the sample times */
int sample_time_comp(const void *i, const void *j)
{
  BSample *d1, *d2;
  d1=(BSample *)i; d2=(BSample *)j;

  return(d1->t - d2->t);
}
/*****************************************************************************/

/*****************************************************************************/
/** Search the date from BData comment lines.
\return Returns 0 when successful.
 */
int bdata_get_date(
  /** Sampler data to be examined */
  BData *bdata,
  /** Pointers to year, month, and day; set to NULL if not needed
   *  (if just checking if date can be found or not) */
  int *yyyy, int *mm, int *dd,
  /** Pointers to hours, min, and sec; set to NULL if not needed;
   *  these may be missing in the header, and that is not considered as error
   */
  int *h, int *m, int *s
) {
  int i, lyyyy=0, lmm=0, ldd=0, lh=0, lm=0, ls=0;
  char *line;
  if(bdata==NULL || bdata->header_nr<1) return 1;
  
  for(i=0; i<bdata->header_nr; i++) {
    line=bdata->header[i];
    if(strlen(line)<12) continue;
    if(line[0]!='#') continue;
    if(line[6]=='-' && line[9]=='-') {
      lyyyy=atoi(line+2); lmm=atoi(line+7); ldd=atoi(line+10);
      break;
    }
  }
  if(lyyyy<1900 || lmm<=0 || lmm>12 || ldd<=0 || ldd>31) return 2;
  if(yyyy!=NULL) *yyyy=lyyyy;
  if(mm!=NULL) *mm=lmm;
  if(dd!=NULL) *dd=ldd;
  /* Try to read time too, if required */
  if(h==NULL && m==NULL && s==NULL) return 0;
  if(h!=NULL) *h=0; if(m!=NULL) *m=0; if(s!=NULL) *s=0;
  if(strlen(line)<21) return 0;
  if(line[15]!=':' || line[18]!=':') return 0;
  lh=atoi(line+13); lm=atoi(line+16); ls=atoi(line+19);
  if(lh<0 || lh>23 || lm<0 || lm>59 || ls<0 || ls>59) return 0;
  if(h!=NULL) *h=lh; if(m!=NULL) *m=lm; if(s!=NULL) *s=ls;
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Count the nr of samples which contain any coincidence counts.
\return Returns 0 when successful.
 */
int bdata_coinc_count(
  /** Sampler data to be examined */
  BData *bdata, 
  /** Nr of samples where channel 1 had coinc counts */
  int *ch1ok,
  /** Nr of samples where channel 2 had coinc counts */
  int *ch2ok,
  /** Nr of samples where both channels had coinc counts */
  int *chbok
) {
  int i, c1nr=0, c2nr=0, bnr=0; 

  if(bdata==NULL) return 1;
  for(i=0; i<bdata->sample_nr; i++) {
    if(bdata->sample[i].pair1_coincid>0) c1nr++;
    if(bdata->sample[i].pair2_coincid>0) c2nr++;
    if(bdata->sample[i].pair1_coincid>0 && bdata->sample[i].pair2_coincid)
      bnr++;
  }
  if(ch1ok!=NULL) *ch1ok=c1nr;
  if(ch2ok!=NULL) *ch2ok=c2nr;
  if(chbok!=NULL) *chbok=bnr;

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Write Scanditronics or GEMS blood online sampler data file.
 *  If file with same name exists, then this function catenates the contents
 *  to the end of it; call first backupExistingFile() if necessary. 
\return Returns 0 when successful. 
 */
int bdata_write(
  /** Pointer to initiated data struct where data will be written. */
  BData *bdata,
  /** Filename for blood sampler data; if file exists, then catenated. */
  char *filename,
  /** Verbose level; set to zero to not to print any comments */
  int verbose,
  /** Pointer to a string (allocated for at least 64 chars) where error message 
   *  or other execution status will be written; enter NULL, if not needed */
  char *status
) {
  FILE *fp;
  int i, n;

  /* Check input */
  if(status!=NULL) strcpy(status, "invalid input");
  if(filename==NULL || strlen(filename)==0) return 1;
  if(bdata==NULL) return 1;
  if(verbose>0) {
    printf("bdata_write(*bdata, %s, %d, status)\n", filename, verbose);
    fflush(stdout);
  }

  /* Open output file */
  if(strcasecmp(filename, "stdout")!=0) {
    if(verbose>1) {printf("opening file %s\n", filename); fflush(stdout);}
    if((fp=fopen(filename, "a")) == NULL) {
      if(status!=NULL) strcpy(status, "cannot open file");
      return 2;
    }
  } else {
    if(verbose>1) {printf("writing into stdout\n"); fflush(stdout);}
    fp=stdout;
  }
  /* Write comment lines */
  if(verbose>1) {printf("writing comment lines\n"); fflush(stdout);}
  for(i=0, n=1; i<bdata->header_nr; i++) {
    n=fprintf(fp, "%s", bdata->header[i]); if(n<1) break;
  }
  if(n<1) {
    if(status!=NULL) strcpy(status, "cannot write in file");
    if(fp!=stdout) fclose(fp); return 3;
  }
  /* Write count data */
  if(verbose>1) {printf("writing data lines\n"); fflush(stdout);}
  for(i=0; i<bdata->sample_nr; i++) {
    n=fprintf(fp, "%9.1f %9.1f %8.1f %7d %7d %7d %7d %7d %7d %7d\n",
        bdata->sample[i].time_of_day,
        bdata->sample[i].time_since_start,
        bdata->sample[i].meas_interval,
        bdata->sample[i].pair1_coincid, bdata->sample[i].pair1_det1_cnts,
        bdata->sample[i].pair1_det2_cnts, bdata->sample[i].pair2_coincid,
        bdata->sample[i].pair2_det1_cnts, bdata->sample[i].pair2_det2_cnts,
        bdata->sample[i].aux_cnts
    );
    if(n<10) break;
  }
  if(fp!=stdout) fclose(fp);
  if(n<10) {
    if(status!=NULL) strcpy(status, "cannot write in file");
    return 4;
  }

  if(verbose>1) {printf("bsampler data written.\n"); fflush(stdout);}
  if(status!=NULL) strcpy(status, "ok");
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Tries to identify sampling start date and time from headers and sample
 *  lines. Time is written in bdata struct as time_t.
\return Returns 0 when successful, otherwise >0.
 */
int bdata_get_start_time(
  /** Pointer to filled blood pump data struct. */
  BData *bdata,
  /** Verbose level; set to zero to not to print any comments */
  int verbose,
  /** Pointer to a string (allocated for at least 64 chars) where error message 
   *  or other execution status will be written; enter NULL, if not needed */
  char *status
) {
  int ret;
  int yyyy, mm, dd, h, m, s;
  time_t t_samp, t_meas;
  struct tm tm_meas={0};
  char temp[256];

  /* Check input */
  if(status!=NULL) strcpy(status, "invalid input");
  if(bdata==NULL || bdata->sample_nr<1) return 1;
  if(verbose>0) {printf("bdata_get_start_time()\n"); fflush(stdout);}

  /* Get start date and time from headers */
  ret=bdata_get_date(bdata, &yyyy, &mm, &dd, &h, &m, &s);
  if(ret!=0) {
    if(status!=NULL) strcpy(status, "date not found in headers");
    if(verbose>0) printf("bdata_get_date := %d\n", ret);
    return 2;
  }
  if(verbose>2) printf("initial_header_time := %04d-%02d-%02d %02d:%02d:%02d\n",
    yyyy, mm, dd, h, m, s);
  tm_meas.tm_year=yyyy-1900; tm_meas.tm_mon=mm-1; tm_meas.tm_mday=dd;
  tm_meas.tm_hour=h; tm_meas.tm_min=m; tm_meas.tm_sec=s; tm_meas.tm_isdst=-1;
  t_meas=mktime(&tm_meas); if(t_meas==(time_t)-1) {
    if(status!=NULL) strcpy(status, "invalid date in headers");
    if(verbose>0) printf("header_time := %04d-%02d-%02d %02d:%02d:%02d\n",
      yyyy, mm, dd, h, m, s);
    return 3;
  }
  if(verbose>1) {
    strftime(temp, 128, "%Y-%m-%d %H:%M:%S", &tm_meas);
    printf("header_time := %s\n", temp);
  }

  /* Get start time (also) from the samples, when necessary */
  if(bdata->sampler_id==BSAMPLER_GEMS) {
    /* GEMS header start time is always wrong (time when sampler was made ready
       but measurement was not yet started); header date is still correct.
       Correct sample times are in sample lines.
     */
    if(verbose>0) printf("GEMS data: date from header, time from samples.\n");
    h=(int)bdata->sample[0].time_of_day;
    /* In case that sample times are from start of 1970 etc */
    h%=86400;
    /* Separate hours, mins and secs from start of day */
    m=h%3600; h-=m; h/=3600; s=m%60; m-=s; m/=60;
    if(verbose>2) printf("initial_first_sample_time := %02d:%02d:%02d\n", h, m, s);
    /* Make t_samp */
    tm_meas.tm_year=yyyy-1900; tm_meas.tm_mon=mm-1; tm_meas.tm_mday=dd;
    tm_meas.tm_hour=h; tm_meas.tm_min=m; tm_meas.tm_sec=s; tm_meas.tm_isdst=-1;
    t_samp=mktime(&tm_meas); if(t_samp==(time_t)-1) {
      if(status!=NULL) strcpy(status, "invalid date or time");
      if(verbose>0) printf("header_time := %04d-%02d-%02d %02d:%02d:%02d\n",
        yyyy, mm, dd, h, m, s);
      return 7;
    }
    if(verbose>1) {
      strftime(temp, 128, "%Y-%m-%d %H:%M:%S", &tm_meas);
      printf("first_sample_time := %s\n", temp);
    }
  } else if(bdata->sampler_id==BSAMPLER_SCANDITRONICS && (h>0 || m>0 || s>0)) {
    /* Scanditronics data where header contains measurement date AND also time */
    if(verbose>0) printf("Scanditronics data: date and time from header.\n");
    t_samp=t_meas;
  } else if(bdata->sampler_id==BSAMPLER_SCANDITRONICS) {
    /* Scanditronics data where sample times must be read from first sample line */
    if(verbose>0)
      printf("Scanditronics data: date from header, time from first sample.\n");
#if(0)
    struct tm *tm_samp;
    tm_samp=localtime((time_t*)&bdata->sample[0].time_of_day);
    if(tm_samp==NULL) {
      if(status!=NULL) strcpy(status, "invalid sample time");
      return 7;
    }
    tm_samp->tm_isdst=-1;
    if(verbose>1) {
      strftime(temp, 128, "%H:%M:%S", tm_samp);
      if(verbose>0) printf("first_sample_time := %s\n", temp);
    }
    /* Make t_samp by combining header date and sample time */
    tm_meas.tm_year=yyyy-1900; tm_meas.tm_mon=mm-1; tm_meas.tm_mday=dd;
    tm_meas.tm_hour=tm_samp->tm_hour; tm_meas.tm_min=tm_samp->tm_min;
    tm_meas.tm_sec=tm_samp->tm_sec; tm_meas.tm_isdst=-1;
    t_samp=mktime(&tm_meas); if(t_samp==(time_t)-1) {
      if(status!=NULL) strcpy(status, "invalid date or time");
      strftime(temp, 128, "%H:%M:%S", tm_samp);
      if(verbose>0) printf("first_sample_time := %04d-%02d-%02d %s\n",
        yyyy, mm, dd, temp);
      return 7;
    }
#else
    h=(int)bdata->sample[0].time_of_day;
    /* Often sample times are from start of 1970, subtract full days */
    h%=86400;
    /* Separate hours, mins and secs from start of day */
    m=h%3600; h-=m; h/=3600; s=m%60; m-=s; m/=60;
    if(verbose>2)
      printf("initial_first_sample_time := %02d:%02d:%02d\n", h, m, s);
    /* Make t_samp by combining header date and sample time */
    tm_meas.tm_year=yyyy-1900; tm_meas.tm_mon=mm-1; tm_meas.tm_mday=dd;
    tm_meas.tm_hour=h; tm_meas.tm_min=m; tm_meas.tm_sec=s; tm_meas.tm_isdst=-1;
    tm_meas.tm_hour+=SAMPLER_GMT_ADD;
    if(tm_meas.tm_hour<0 || tm_meas.tm_hour>23) tm_meas.tm_hour-=SAMPLER_GMT_ADD; 
    t_samp=mktime(&tm_meas); if(t_samp==(time_t)-1) {
      if(status!=NULL) strcpy(status, "invalid date or time");
      if(verbose>0) printf("sample_time := %04d-%02d-%02d %02d:%02d:%02d\n",
        yyyy, mm, dd, h, m, s);
      return 7;
    }
#endif
    if(verbose>1) {
      strftime(temp, 128, "%Y-%m-%d %H:%M:%S", &tm_meas);
      printf("first_sample_time := %s\n", temp);
    }
  } else {
    if(status!=NULL) strcpy(status, "sampler not supported");
    if(verbose>0) printf("sampler %d not supported here\n", bdata->sampler_id);
    return 9;
  }

  bdata->start_time=t_samp;

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Correct Blood sampler header based on contents.
\return Retuens 0 when successful.
 */
int bdata_correct_header(
  /** Pointer to filled blood pump data struct */
  BData *bdata
) {
  int i, ret;
  char *line, temp[128];
  struct tm *tm_start;

  /* Check input */
  if(bdata==NULL) return 1;
  if(bdata->start_time==(time_t)-1) return 1;

  /* Create new header line for measurement start date and time */
  tm_start=localtime(&bdata->start_time); if(tm_start==NULL) return 2;
  tm_start->tm_isdst=-1;
  strftime(temp, 128, "# %Y-%m-%d %H:%M:%S\n", tm_start);

  /* Search the previous header line fr date and time */
  for(i=0, ret=1; i<bdata->header_nr; i++) {
    line=bdata->header[i];
    if(strlen(line)<12) continue;
    if(line[0]!='#') continue;
    if(line[6]=='-' && line[9]=='-') {
      strcpy(bdata->header[i], temp);
      ret=0; break;
    }
  }
  if(ret!=0) return 3;

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
