/** @file simdispersion.c
 *  @brief Simulation of dispersion.
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpccm.h"
/*****************************************************************************/

/*****************************************************************************/
/** Simulate the effect of dispersion on a time-activity curve.
 *   
 *  @details
 *  Memory for y must be allocated in the calling program.
 *  The units of rate constants must be related to the time unit;
 *  1/min and min, or 1/sec and sec.
 * 
 *  @return Function returns 0 when successful, else a value >= 1.
 *  @author Vesa Oikonen
 */
int simDispersion(
  /** Array of sample times */
  double *x,
  /** Array of sample values, which will be replaced here by dispersion added
   *  values */
  double *y,
  /** Nr of samples */
  const int n,
  /** First dispersion time constant (zero if no dispersion); in same
   *  time unit as sample times */ 
  const double tau1,
  /** 2nd dispersion time constant (zero if no dispersion); in same
   *  time unit as sample times */ 
  const double tau2,
  /** Array for temporary data, for at least n samples */
  double *tmp
) {
  int i, ret;
  double k;

  /* Check input */
  if(x==NULL || y==NULL || tmp==NULL || n<2) return 1;
  if(tau1<0.0 || tau2<0.0) return 2;

  /* First dispersion */
  if(tau1>0.0) {
    k=1.0/tau1;
    ret=simC1(x, y, n, k, k, tmp); if(ret!=0) return 100+ret;
    for(i=0; i<n; i++) y[i]=tmp[i];
  }

  /* Second dispersion */
  if(tau2>0.0) {
    k=1.0/tau2;
    ret=simC1(x, y, n, k, k, tmp); if(ret!=0) return 200+ret;
    for(i=0; i<n; i++) y[i]=tmp[i];
  }

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
