/** @file tpcift.h
 *  @brief Header file for library libtpcift.
 *  @details Interfile-type header information processing and file i/o
 *  @author Vesa Oikonen
 */
#ifndef _TPCIFT_H_
#define _TPCIFT_H_
/*****************************************************************************/

/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/*****************************************************************************/
#include "tpcextensions.h"
/*****************************************************************************/

/*****************************************************************************/
/** Struct for one IFT item containing key and its value. Add new item using iftPut().
   @sa IFT, iftPut
 */
typedef struct IFT_ITEM {
  /** Comment line, starting with '#': 0=not, 1=yes. */
  char comment;
  /** Undefined short integer for the use of programmer. */
  short int sw;
  /** Pointer to the NULL terminated key string; NULL if not allocated;
      Set or change this only using dedicated functions. */
  char *key;
  /** Pointer to the NULL terminated key value string; NULL if not allocated.
      Set or change this only using dedicated functions.
      If this field is used to store decimal number or list of decimal numbers, then 
      do NOT use decimal commas. */
  char *value;
} IFT_ITEM;

/** Main struct for IFT items.
   @sa IFT_ITEM, iftInit, iftFree, CVS, TAC, PAR
 */
typedef struct IFT {
  /** Number of allocated places for keys and values */
  int _memNr;
  /** Number of stored keys and their values. */
  int keyNr;
  /** Type of the parameter file:
      0=unknown, 1=interfile ':=' , 2=setup '=', 3=result ':', 4=space ' ', 5=tab, 6=',', 7=';'
  */
  int type;
  /** Space before equal sign (0=no, otherways yes). */
  int space_before_eq;
  /** Space after equal sign (0=no, otherways yes). */
  int space_after_eq;
  /** List of key-value -pairs. */
  IFT_ITEM *item;
} IFT;
/*****************************************************************************/

/*****************************************************************************/
/* ift */
void iftInit(IFT *ift);
void iftFree(IFT *ift);
int iftPut(IFT *ift, const char *key, const char *value, char comment, TPCSTATUS *status);
int iftPutDouble(IFT *ift, const char *key, const double value, char comment, TPCSTATUS *status);
int iftPutInt(IFT *ift, const char *key, const int value, char comment, TPCSTATUS *status);
int iftPutUInt(IFT *ift, const char *key, const unsigned int value, char comment, TPCSTATUS *status);
int iftDelete(IFT *ift, int index);
int iftDuplicate(IFT *ift1, IFT *ift2);
int iftReplaceValue(IFT *ift, int i, const char *value, TPCSTATUS *status);
int iftReplaceKey(IFT *ift, int i, const char *key, TPCSTATUS *status);
int iftDeleteDuplicateKeys(IFT *ift, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* iftio */
int iftWrite(IFT *ift, FILE *fp, TPCSTATUS *status);
int iftWriteItem(IFT *ift, int item, FILE *fp, TPCSTATUS *status);
int iftRead(IFT *ift, FILE *fp, int is_key_required, int is_comment_accepted, TPCSTATUS *status);
int iftPutFromString(
  IFT *ift, const char *line, int is_key_required, int is_comment_accepted, TPCSTATUS *status
);
int iftPutFromStringWithSpaceSeparator(
  IFT *ift, const char *line, int is_comment_accepted, TPCSTATUS *status
);
/*****************************************************************************/

/*****************************************************************************/
/* iftfind */
int iftFindKey(IFT *ift, const char *key, int start_index);
int iftFindPair(IFT *ift, const char *key, const char *value, int start_index);
int iftSearchKey(IFT *ift, const char *s, int start_index);
int iftSearchValue(IFT *ift, const char *s, int start_index);
int iftFindNrOfKeys(IFT *ift, const char *key);
void iftDeleteKey(IFT *ift, const char *key);
int iftGetDoubleValue(IFT *ift, const char *key, int index, double *v);
/*****************************************************************************/

/*****************************************************************************/
/* iftget */
int iftGetFloat(IFT *ift, int index, float *v);
int iftGetDouble(IFT *ift, int index, double *v);
int iftGetInt(IFT *ift, int index, int *v);
int iftGetUInt(IFT *ift, int index, unsigned int *v);
int iftGetDoubleWithUnit(IFT *ift, int index, double *v, int *unit);
/*****************************************************************************/

/*****************************************************************************/
#endif /* TPCIFT */
