/** @file simplex.c
 *  @brief Downhill simplex nonlinear optimization.
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpcextensions.h"
/*****************************************************************************/
#include "tpcnlopt.h"
/*****************************************************************************/

/*****************************************************************************/
/** Nelder-Mead (downhill simplex) optimization. 
    Number of free parameters (not fixed) must be > 1.
    @pre Initiate the contents of the nlo data structure.
    @post The last objective function call is usually not done with the best parameter estimates;
    if objective function simulates data that you need, you must call the function with the final
    parameters.
    @return enum tpcerror (TPCERROR_OK when successful).
    @author Vesa Oikonen
 */
int nloptSimplex(
  /** Pointer to NLOPT struct. 
      @pre Initial guess must be given in x[].
      Initial step sizes must be given in xdelta[]. 
      Constraints xlower[] and xupper[] are required, but since the
      implementation is very simplified, limits should be as wide as possible. 
      Parameter tolerances are used as stopping criteria: Simplex stops if centroid differs from 
      best point less than xtol[], for each parameter; if set to zero, then only stopping rule is 
      the iteration number.
   */
  NLOPT *nlo,
  /** Maximum number of iterations; set to zero to use the default. */
  unsigned int maxIter,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s(NLOPT, %d, status)\n", __func__, maxIter);
  if(nlo==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_FAIL);
    return TPCERROR_FAIL;
  }
  if(nlo->totalNr<1 || nlo->xfull==NULL || nlo->_fun==NULL) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }

  unsigned int parNr=nlo->totalNr;
  if(parNr<2) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_PARNR);
    return TPCERROR_INVALID_PARNR;
  }

  /* Check that initial values are inside limits */
  for(unsigned int i=0; i<nlo->totalNr; i++)
    if(nlo->xfull[i]<nlo->xlower[i] || nlo->xfull[i]>nlo->xupper[i]) {
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_VALUE);
      return TPCERROR_INVALID_VALUE;
    }


  /* Initiate the simplex */
  if(verbose>10) printf("Simplex initialization\n");
  unsigned int sn=parNr+1; // sn>=3 because parNr >=2 verified before
  double simplexP[sn+2][parNr]; // including room for first and second new point
  double simplexC[parNr];
  double simplexR[sn+2]; // including room for first and second new point
  if(verbose>16) printf("simplexR[0..%d]\n", sn+2);
  unsigned int newPnt=sn;
  unsigned int newPnt2=newPnt+1;
  /* Check the maximum iteration number */
  if(maxIter<1) maxIter=500;

  /* Fill with initial guesses */
  for(unsigned int j=0; j<sn; j++)
    for(unsigned int i=0; i<parNr; i++)
      simplexP[j][i]=nlo->xfull[i];

  for(unsigned int j=1; j<sn; j++) { // initial guess is kept as such in initial simplex
    for(unsigned int i=0; i<parNr; i++) {
      simplexP[j][i] = simplexP[j-1][i];
      if((j-1)==i) {
        simplexP[j][i] += nlo->xdelta[i];
        if(simplexP[j][i]<nlo->xlower[i] || simplexP[j][i]>nlo->xupper[i]) {
          /* parameter would exceed limits, try the other direction */
          simplexP[j][i] -= nlo->xdelta[i];
          if(simplexP[j][i]<nlo->xlower[i] || simplexP[j][i]>nlo->xupper[i]) {
            /* stupid limits or delta */
            statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_VALUE);
            return TPCERROR_INVALID_VALUE;
          }
        }
      }
    }
  }
  /* Calculate initial simplex function values */
  for(unsigned int j=0; j<sn; j++) {
    if(verbose>13) {
      printf("x[%d][]=", j);
      for(unsigned int i=0; i<parNr; i++) printf("%g ", simplexP[j][i]);
      printf("\n");
    }
    simplexR[j] = (*nlo->_fun)(parNr, simplexP[j], nlo->fundata);
    nlo->funCalls++;
    if(verbose>12) printf("  R[%d]=%g\n", j, simplexR[j]);
    if(nlo->usePList) {
      int ret=nloptAddP(nlo, simplexP[j], simplexR[j]);
      if(ret!=TPCERROR_OK) {
        statusSet(status, __func__, __FILE__, __LINE__, ret);
        return(ret);
      }
    }
  }
  double initialR=simplexR[0];
#if(0)
  double lastChi=simplexR[0]; // used in loop to check whether proceeding
  for(unsigned int j=1; j<sn; j++) if(lastChi<simplexR[j]) lastChi=simplexR[j];
#endif

  /* Simplex iterations */
  if(verbose>10) printf("Simplex iterations\n");
  unsigned int iterNr=0;
  unsigned int best=0, nextBest=0, worst=0;
  double max, min, min2;
  double f;
  /* Find the max and min response measured */
  max=min=simplexR[0]; best=worst=0; // initial guess
  for(unsigned int j=1; j<sn; j++) {
    if(!isfinite(max) || simplexR[j] > max) {max=simplexR[j]; worst=j;}
    if(!isfinite(min) || simplexR[j] < min) {min=simplexR[j]; best=j; }
  }
  /* Find the 2nd best, too */
  min2=max; nextBest=worst;
  for(unsigned int j=0; j<sn; j++)
    if(j!=best && (simplexR[j] < min2)) {min2=simplexR[j]; nextBest=j;}
  if(verbose>14) {
    printf("  best := %g (%d)\n", min, best);
    printf("  next best := %g (%d)\n", min2, nextBest);
    printf("  worst := %g (%d)\n", max, worst);
  }
  do {
    if(verbose>13) printf("\niteration := %d\n", 1+iterNr);
    /* Calculate centroid of all measurements except the worst */
    for(unsigned int i=0; i<parNr; i++) {
      simplexC[i]=0.0;
      for(unsigned int j=0; j<sn; j++) if(j!=worst) simplexC[i]+=simplexP[j][i];
      simplexC[i]/=(double)(sn-1);
    }
    if(verbose>15) {
      printf("centroid x[]=");
      for(unsigned int i=0; i<parNr; i++) printf("%g ", simplexC[i]);
      printf("\n");
    }
    /* Stopping rule: If centroid and best point differ less than 0.1*tolerance for each parameter, 
       then quit iterations */
    unsigned int k;
    for(k=0; k<parNr; k++) if(fabs(simplexC[k]-simplexP[best][k])>0.1*nlo->xtol[k]) break;
    if(k==parNr) break; // no break in previous loop - exit the do-while loop
    /* Create new point as centroid reflected away from worst */
    f=1.0;
    for(unsigned int i=0; i<parNr; i++) {
      simplexP[newPnt][i] = simplexC[i] + f*(simplexC[i]-simplexP[worst][i]);
      /* Make sure it stays within limits */
      if(simplexP[newPnt][i]<nlo->xlower[i]) simplexP[newPnt][i]=nlo->xlower[i];
      else if(simplexP[newPnt][i]>nlo->xupper[i]) simplexP[newPnt][i]=nlo->xupper[i];
    }
    /* and compute the response */
    simplexR[newPnt] = (*nlo->_fun)(parNr, simplexP[newPnt], nlo->fundata);
    nlo->funCalls++;
    if(verbose>15) {
      printf("new point x[%d][]=", newPnt);
      for(unsigned int i=0; i<parNr; i++) printf("%g ", simplexP[newPnt][i]);
      printf("\n  -> R[%d]=%g\n", newPnt, simplexR[newPnt]);
    }
    if(nlo->usePList) {
      int ret=nloptAddP(nlo, simplexP[newPnt], simplexR[newPnt]);
      if(ret!=TPCERROR_OK) {
        statusSet(status, __func__, __FILE__, __LINE__, ret);
        return(ret);
      }
    }
    /* Depending on the response, decide what to do next */
    if(simplexR[newPnt] < simplexR[best]) {
      if(verbose>14) printf("  new is better than best\n");
      /* If this new point is better than previous best, then expand in this direction */
      f=2.0;
      for(unsigned int i=0; i<parNr; i++) {
        simplexP[newPnt2][i] = simplexC[i] + f*(simplexC[i]-simplexP[worst][i]);
        //simplexP[newPnt2][i] = simplexC[i] + f*(simplexP[newPnt][i]-simplexC[i]);
        /* Make sure it stays within limits */
        if(simplexP[newPnt2][i]<nlo->xlower[i]) simplexP[newPnt2][i]=nlo->xlower[i];
        else if(simplexP[newPnt2][i]>nlo->xupper[i]) simplexP[newPnt2][i]=nlo->xupper[i];
      }
      simplexR[newPnt2] = (*nlo->_fun)(parNr, simplexP[newPnt2], nlo->fundata);
      nlo->funCalls++;
      if(nlo->usePList) {
        int ret=nloptAddP(nlo, simplexP[newPnt2], simplexR[newPnt2]);
        if(ret!=TPCERROR_OK) {
          statusSet(status, __func__, __FILE__, __LINE__, ret);
          return(ret);
        }
      }
      if(simplexR[newPnt2] < simplexR[newPnt]) {
        for(unsigned int i=0; i<parNr; i++) simplexP[worst][i]=simplexP[newPnt2][i];
        simplexR[worst]=simplexR[newPnt2];
      } else {
        for(unsigned int i=0; i<parNr; i++) simplexP[worst][i]=simplexP[newPnt][i];
        simplexR[worst]=simplexR[newPnt];
      }
    } else if(!isfinite(simplexR[newPnt]) || simplexR[newPnt] > simplexR[worst]) {
      if(verbose>14) printf("  new is worse than worst\n");
      /* If new point is worse than previous worst, measure point halfway between 
         the worst and centroid */
      f=-0.5;
      for(unsigned int i=0; i<parNr; i++) {
        simplexP[newPnt2][i] = simplexC[i] + f*(simplexC[i]-simplexP[worst][i]);
        /* Make sure it stays within limits */
        if(simplexP[newPnt2][i]<nlo->xlower[i]) simplexP[newPnt2][i]=nlo->xlower[i];
        else if(simplexP[newPnt2][i]>nlo->xupper[i]) simplexP[newPnt2][i]=nlo->xupper[i];
      }
      simplexR[newPnt2] = (*nlo->_fun)(parNr, simplexP[newPnt2], nlo->fundata);
      nlo->funCalls++;
      if(nlo->usePList) {
        int ret=nloptAddP(nlo, simplexP[newPnt2], simplexR[newPnt2]);
        if(ret!=TPCERROR_OK) {
          statusSet(status, __func__, __FILE__, __LINE__, ret);
          return(ret);
        }
      }
      if(simplexR[newPnt2] < simplexR[newPnt] && isfinite(simplexR[newPnt2])) {
        for(unsigned int i=0; i<parNr; i++) simplexP[worst][i]=simplexP[newPnt2][i];
        simplexR[worst]=simplexR[newPnt2];
      } else if(isfinite(simplexR[newPnt])) {
        for(unsigned int i=0; i<parNr; i++) simplexP[worst][i]=simplexP[newPnt][i];
        simplexR[worst]=simplexR[newPnt];
      }
    } else if(simplexR[newPnt] > simplexR[nextBest]) {
      if(verbose>14) printf("  new is worse than next best\n");
      /* If newest response is worse than next best point but better than worst, 
         measure response halfway between centroid and the newest point */
      f=0.5;
      for(unsigned int i=0; i<parNr; i++) {
        simplexP[newPnt2][i] = simplexC[i] + f*(simplexP[newPnt][i]-simplexC[i]);
        /* Make sure it stays within limits */
        if(simplexP[newPnt2][i]<nlo->xlower[i]) simplexP[newPnt2][i]=nlo->xlower[i];
        else if(simplexP[newPnt2][i]>nlo->xupper[i]) simplexP[newPnt2][i]=nlo->xupper[i];
      }
      simplexR[newPnt2] = (*nlo->_fun)(parNr, simplexP[newPnt2], nlo->fundata);
      nlo->funCalls++;
      if(nlo->usePList) {
        int ret=nloptAddP(nlo, simplexP[newPnt2], simplexR[newPnt2]);
        if(ret!=TPCERROR_OK) {
          statusSet(status, __func__, __FILE__, __LINE__, ret);
          return(ret);
        }
      }
      if(simplexR[newPnt2] < simplexR[newPnt]) {
        if(verbose>14) printf("  halfway is better than new\n");
        for(unsigned int i=0; i<parNr; i++) simplexP[worst][i]=simplexP[newPnt2][i];    
        simplexR[worst]=simplexR[newPnt2];
      } else if(isfinite(simplexR[newPnt])) {
        if(verbose>14) printf("  halfway is worse than new\n");
        for(unsigned int i=0; i<parNr; i++) simplexP[worst][i]=simplexP[newPnt][i];
        simplexR[worst]=simplexR[newPnt];
      }
    } else if(isfinite(simplexR[newPnt])) {
      if(verbose>14) printf("  replace the worst with this\n");
      /* If none of the above is true, then replace the worst point with this, for now */
      for(unsigned int i=0; i<parNr; i++) simplexP[worst][i]=simplexP[newPnt][i];
      simplexR[worst]=simplexR[newPnt];
    }

    /* Find the max and min response measured */
    max=min=simplexR[0]; best=worst=0;
    for(unsigned int j=1; j<sn; j++) {
      if(!isfinite(max) || simplexR[j] > max) {max=simplexR[j]; worst=j;}
      if(!isfinite(min) || simplexR[j] < min) {min=simplexR[j]; best=j; }
    }
    /* Find the 2nd best, too */
    min2=max; nextBest=worst;
    for(unsigned int j=0; j<sn; j++)
      if(j!=best && (simplexR[j] < min2)) {min2=simplexR[j]; nextBest=j;}
    if(verbose>14) {
      printf("  best := %g (%d)\n", min, best);
      printf("  next best := %g (%d)\n", min2, nextBest);
      printf("  worst := %g (%d)\n", max, worst);
    }

    iterNr++;
  } while(iterNr<maxIter);
  if(verbose>2) printf("  %d iterations\n", iterNr);
  if(verbose>2) printf("  R[%d]=%g\n", best, simplexR[best]);

#if(0)
  /* One more function call with the best parameters, to make sure that
     if objective function simulates data, it is simulated with the best fit */
   simplexR[best] = (*nlo->_fun)(parNr, simplexP[best], nlo->fundata);
   nlo->funCalls++;
#endif

  /* Check the objective function value */
  if(!isfinite(simplexR[best])) return(TPCERROR_BAD_FIT);

  /* Copy optimized parameters over initial values */
  if(simplexR[best]<=initialR) {
    for(unsigned int i=0; i<parNr; i++) nlo->xfull[i]=simplexP[best][i];
    nlo->funval=simplexR[best];
  } else {
    /* This happens only in case of a bug */
    if(1 || verbose>0) printf("nloptSimplex() gives worse R than initially!\n");
    return(TPCERROR_BAD_FIT);
  }

  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
