/** @file tpcnlopt.h
 *  @brief Header file for library libtpcnlopt.
 *  @details Non-linear optimization.
 *  @author Vesa Oikonen
 */
#ifndef _TPCNLOPT_H_
#define _TPCNLOPT_H_
/*****************************************************************************/

/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/*****************************************************************************/
#include "tpcextensions.h"
#include "tpcrand.h"
/*****************************************************************************/

/*****************************************************************************/
/** General settings for NLOPT routines; individual routines may not use all of these,
    and can require additional parameters. */
typedef struct NLOPT {
  /** Total nr of parameters (dimensions). */
  unsigned int totalNr;
  /** Array of all parameter values, including fixed ones. */
  double       *xfull;
  /** Array of lower limits for parameters, including fixed ones. */
  double       *xlower;
  /** Array of upper limits for parameters, including fixed ones. */
  double       *xupper;
  /** Array of (initial) steps for parameters, including fixed ones;
      set it to the expected distance from the solution. */
  double       *xdelta;
  /** Array of tolerances for parameters, including fixed ones; may be used
      as a stopping rule and for scaling. */
  double       *xtol;
  /** Pointer to objective function; parameters are optimized to minimize
      the return value of this function. */
  double (*_fun)(int, double*, void*);
  /** Pointer to data which is passed on to the function; NULL if not needed. */
  void *fundata;
  /** Maximum allowed number of function calls; set to zero for no limit. */
  unsigned int maxFunCalls;
  /** Number of function calls so far. */
  unsigned int funCalls;
  /** The return value of objective function with xfull[] and fundata. */
  double funval;
  /** Keep (1) or do not keep (0) plist (below). */
  unsigned int usePList;
  /** List of all visited parameters and objective function values;
      stored in order x[], funval, x[], funval, ... with length funCalls*(totalNr+1);
      updated only if usePList!=0 (above). */
  double       *plist;
} NLOPT;
/*****************************************************************************/

/*****************************************************************************/
/* nlopt */
void nloptInit(NLOPT *nlo);
void nloptFree(NLOPT *nlo);
int nloptAllocate(NLOPT *nlo, unsigned int parNr);
int nloptDuplicate(NLOPT *nlo1, NLOPT *nlo2);
int nloptAddP(NLOPT *nlo, double *p, double funval);
void nloptWrite(NLOPT *d, FILE *fp);
unsigned int nloptLimitFixedNr(NLOPT *d);
unsigned int nloptFixedNr(NLOPT *d);
void nloptRemoveEmpties(NLOPT *d);
/*****************************************************************************/

/*****************************************************************************/
/* rndpoint */
int nloptRandomPoint(double *p, double *low, double *up, unsigned int n);
/*****************************************************************************/

/*****************************************************************************/
/* constraints */
unsigned int nloptCheckParameters(
  unsigned int n, double *pLower, double *pUpper, double *p,
  double *pAccept, double *penalty
);
/*****************************************************************************/

/*****************************************************************************/
/* simplex */
int nloptSimplex(NLOPT *nlo, unsigned int maxIter, TPCSTATUS *);
/*****************************************************************************/

/*****************************************************************************/
/* praxis */
int nloptPowellBrent(NLOPT *nlo, unsigned int ktm, TPCSTATUS *);
/*****************************************************************************/

/*****************************************************************************/
/* mpso */
int nloptMPSO(NLOPT *nlo, const int doLocal, TPCSTATUS *);
/*****************************************************************************/

/*****************************************************************************/
/* tgo */
int nloptITGO1(
  NLOPT *nlo, const int doLocal, unsigned int tgoNr, unsigned int sampleNr,
  unsigned int neighNr, TPCSTATUS *status
);
int nloptITGO2(
  NLOPT *nlo, const int doLocal, unsigned int tgoNr, unsigned int sampleNr,
  unsigned int neighNr, TPCSTATUS *status
);
int nloptIATGO(NLOPT *nlo, const int doLocal, unsigned int maxIterNr,
  double neighFract, TPCSTATUS *status
);
/*****************************************************************************/

/*****************************************************************************/
#endif /* TPCNLOPT */
