/** @file shuffle.c
 *  @brief Random shuffle and related functions.
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpcextensions.h"
/*****************************************************************************/
#include "tpcrand.h"
/*****************************************************************************/

/*****************************************************************************/
/** @brief Random shuffle
 *
 *  Arrange the n elements of integer array in random order.
 *  Only effective if N is much smaller than UINT64_MAX.
 *  
 *  @pre Prepare the Mersenne Twister, for example using code
 *       MERTWI mt; mertwiInit(&mt); mertwiInitWithSeed64(&mt, mertwiSeed64());
 *  @sa randomShuffleUnsigned
 */
void randomShuffle(
  /** Integer array to be shuffled */
  int *array, 
  /** Array length */
  unsigned int n,
  /** Data struct for Mersenne Twister MT19937 pseudorandom number generator */
  MERTWI *mt
) {
  if(n<=1 || array==NULL) return;
  unsigned int i, j;
  int tmp;
  for(i=0; i<n-1; i++) {
    j=i+mertwiRandomInt64(mt)/(UINT64_MAX/(n-i)+1);
    tmp=array[j]; array[j]=array[i]; array[i]=tmp;
  }
}
/*****************************************************************************/

/*****************************************************************************/
/** @brief Random shuffle
 *
 *  Arrange the n elements of unsigned integer array in random order.
 *  Only effective if N is much smaller than UINT64_MAX.
 *  
 *  @pre Prepare the Mersenne Twister, for example using code
 *       MERTWI mt; mertwiInit(&mt); mertwiInitWithSeed64(&mt, mertwiSeed64());
 *  @sa randomShuffle
 *  
 */
void randomShuffleUnsigned(
  /** Integer array to be shuffled */
  unsigned int *array, 
  /** Array length */
  unsigned int n,
  /** Data struct for Mersenne Twister MT19937 pseudorandom number generator */
  MERTWI *mt
) {
  if(n<=1 || array==NULL) return;
  unsigned int i, j, tmp;
  for(i=0; i<n-1; i++) {
    j=i+mertwiRandomInt64(mt)/(UINT64_MAX/(n-i)+1);
    tmp=array[j]; array[j]=array[i]; array[i]=tmp;
  }
}
/*****************************************************************************/

/*****************************************************************************/
/** @brief Random permutation for an array of signed integers.
 *
 *  Given (allocated) array of length n is filled with random permutation of 
 *  numbers (signed integers) in the range [a:n+a-1]; that is, each number
 *  once in random order.
 *  @pre Prepare the Mersenne Twister, for example using code
 *       MERTWI mt; mertwiInit(&mt); mertwiInitWithSeed64(&mt, mertwiSeed64());
 *  @sa randomPermutationUnsigned
 */
void randomPermutation(
  /** Allocated array to be filled */
  int *array, 
  /** Length of array */
  unsigned int n,
  /** Smallest integer value */ 
  int a,
  /** Data struct for Mersenne Twister MT19937 pseudorandom number generator */
  MERTWI *mt
) {
  if(n<1 || array==NULL) return;
  int i;
  for(i=0; i<(int)n; i++) array[i]=i+a;
  randomShuffle(array, n, mt);
}
/*****************************************************************************/

/*****************************************************************************/
/** @brief Random permutation for an array of unsigned integers.
 *
 *  Given (allocated) array of length n is filled with random permutation of 
 *  numbers (unsigned integers) in the range [a:n+a-1]; that is, each number
 *  once in random order.
 *  @pre Prepare the Mersenne Twister, for example using code
 *       MERTWI mt; mertwiInit(&mt); mertwiInitWithSeed64(&mt, mertwiSeed64());
 *  @sa randomPermutation
 */
void randomPermutationUnsigned(
  /** Allocated array to be filled */
  unsigned int *array, 
  /** Length of array */
  unsigned int n,
  /** Smallest integer value */ 
  unsigned int a,
  /** Data struct for Mersenne Twister MT19937 pseudorandom number generator */
  MERTWI *mt
) {
  if(n<1 || array==NULL) return;
  unsigned int i;
  for(i=0; i<n; i++) array[i]=i+a;
  randomShuffleUnsigned(array, n, mt);
}
/*****************************************************************************/

/*****************************************************************************/

#if(0)
/** @brief Random shuffle
 *
 *  Arrange the n elements of integer array in random order.
 *  Only effective if N is much smaller than RAND_MAX.
 *  
 *  @pre Uses rand(), therefore set seed for a new series of pseudo-random 
 *  numbers; to produce truly random numbers (not just pseudo-random), 
 *  do srand(time(0)) before calling this function. If no seed is set, then 
 *  value 1 is used as default seed by rand().
 *  @sa randomShuffleUnsigned
 */
void randomShuffle(
  /** Integer array to be shuffled */
  int *array, 
  /** Array length */
  unsigned int n
) {
  if(n<=1 || array==NULL) return;
  unsigned int i, j;
  int tmp;
  for(i=0; i<n-1; i++) {
    j=i+rand()/(RAND_MAX/(n-i)+1);
    tmp=array[j]; array[j]=array[i]; array[i]=tmp;
  }
}

/** @brief Random shuffle
 *
 *  Arrange the n elements of unsigned integer array in random order.
 *  Only effective if N is much smaller than RAND_MAX.
 *  
 *  @pre Uses rand(), therefore set seed for a new series of pseudo-random 
 *  numbers; to produce truly random numbers (not just pseudo-random), 
 *  do srand(time(0)) before calling this function. If no seed is set, then 
 *  value 1 is used as default seed by rand().
 *  @sa randomShuffle
 *  
 */
void randomShuffleUnsigned(
  /** Integer array to be shuffled */
  unsigned int *array, 
  /** Array length */
  unsigned int n
) {
  if(n<=1 || array==NULL) return;
  unsigned int i, j, tmp;
  for(i=0; i<n-1; i++) {
    j=i+rand()/(RAND_MAX/(n-i)+1);
    tmp=array[j]; array[j]=array[i]; array[i]=tmp;
  }
}

/** @brief Random permutation.
 *
 *  Given (allocated) array of length n is filled with random permutation of 
 *  numbers (signed integers) in the range [a:n+a-1]; that is, each number
 *  once in random order.
 *  @sa randomPermutationUnsigned
 */
void randomPermutation(
  /** Allocated array to be filled */
  int *array, 
  /** Length of array */
  unsigned int n,
  /** Smallest integer value */ 
  int a
) {
  if(n<1 || array==NULL) return;
  int i;
  for(i=0; i<(int)n; i++) array[i]=i+a;
  randomShuffle(array, n);
}

/** @brief Random permutation.
 *
 *  Given (allocated) array of length n is filled with random permutation of 
 *  numbers (unsigned integers) in the range [a:n+a-1]; that is, each number
 *  once in random order.
 *  @sa randomPermutation
 */
void randomPermutationUnsigned(
  /** Allocated array to be filled */
  unsigned int *array, 
  /** Length of array */
  unsigned int n,
  /** Smallest integer value */ 
  unsigned int a
) {
  if(n<1 || array==NULL) return;
  unsigned int i;
  for(i=0; i<n; i++) array[i]=i+a;
  randomShuffleUnsigned(array, n);
}
/*****************************************************************************/

/*****************************************************************************/
#endif
