/** @file tacw.c
 *  @brief Working with weights in TAC struct.
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include "tpcift.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpctac.h"
/*****************************************************************************/

/*****************************************************************************/
/** Copy weights from one TAC struct to another.
 *  @sa tacXCopy, tacWMove
 *  @author Vesa Oikonen
 *  @return Returns TPCERROR status.
 */
int tacWCopy(
  /** Pointer to source TAC struct */
  TAC *tac1,
  /** Pointer to target TAC struct */
  TAC *tac2,
  /** Index of the first sample to copy */
  int i1,
  /** Index of the last sample to copy */
  int i2
) {
  if(tac1==NULL || tac2==NULL) return(TPCERROR_FAIL);
  if(i1<0 || i1>i2) return(TPCERROR_FAIL);
  if(i2>=tac1->_sampleNr || i2>=tac2->_sampleNr) return(TPCERROR_FAIL);
  
  for(int i=i1; i<=i2; i++) {
    tac2->w[i]=tac1->w[i];
  }
  tac2->weighting=tac1->weighting;
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
/** Identify column containing weights in TAC struct, and move weights to
 *  the correct place. 
 *  
 *  TAC name starting with string 'weight' is identified as weight column.
 *  
 *  Sets tac->weighting to WEIGHTING_ON or WEIGHTING_OFF, depending on whether
 *  weight column was found or not.
 *
 *  @sa tacWCopy
 *  @author Vesa Oikonen
 *  @return enum tpcerror (TPCERROR_OK when successful).
 */
int tacWMove(
  /** Pointer to TAC struct */
  TAC *tac,
  /** Overwrite (1) or do not overwrite (0) existing weights. */
  int ow,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  if(tac==NULL || tac->tacNr<1 || tac->sampleNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  int ret;

  if(tac->weighting==WEIGHTING_UNKNOWN) tac->weighting=WEIGHTING_OFF;

  /* Search for the weight column */
  int wc=-1;
  for(int i=0; i<tac->tacNr; i++) {
    if(strncasecmp(tac->c[i].name, "WEIGHT", 6)!=0) continue;
    if(wc<0) wc=i;
    else if(ow==0) {
      /* Error if found more than one */
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_DUPLICATE_DATA);
      return TPCERROR_DUPLICATE_DATA;
    }
  }
  /* Error if not found */
  if(wc<0) {
    if(ow) tac->weighting=WEIGHTING_OFF;
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_WEIGHTS);
    return TPCERROR_NO_WEIGHTS;
  }
  /* Error if overwriting not allowed */
  if(ow==0 && tac->weighting==WEIGHTING_ON) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_DUPLICATE_DATA);
    return TPCERROR_DUPLICATE_DATA;
  }
  /* Copy weights to their right place */
  for(int i=0; i<tac->sampleNr; i++) tac->w[i]=tac->c[wc].y[i];
  tac->weighting=WEIGHTING_ON;
  /* Then delete the original column */
  ret=tacDeleteTACC(tac, wc);
  statusSet(status, __func__, __FILE__, __LINE__, ret);
  return(ret);
}
/*****************************************************************************/

/*****************************************************************************/
/** Add weight to TAC data based on sample frequency or time frame length.

    Weights are scaled so that weight sum = sample number.
    Any existing weighting is overwritten.

    @sa tacWSampleNr, tacWCopy, tacWMove
    @return Return enum tpcerror (TPCERROR_OK when successful).
 */
int tacWByFreq(
  /** Pointer to TAC struct where weights will be added.
      Samples/frames must be sorted by sample time, but duplicate samples
      are allowed */
  TAC *tac,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  if(tac==NULL || tac->tacNr<1 || tac->sampleNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  
  tac->weighting=WEIGHTING_ON;
  if(tac->sampleNr==1) {tac->w[0]=1.0; return(TPCERROR_OK);}

  double sumw=0.0;
  if(tac->isframe!=0) { /* weights based on frame lengths */

    for(int i=0; i<tac->sampleNr; i++) {
      tac->w[i]=tac->x2[i]-tac->x1[i]; 
      sumw+=tac->w[i];
    }

  } else { /* weights based on sample distance */

    int i, i1, i2;
    double t, t1, t2, f;
    for(i=0; i<tac->sampleNr; i++) {
      t=t1=t2=tac->x[i];
      /* Find the closest sample time before this one */
      for(i1=i; i1>=0; i1--) {t1=tac->x[i1]; if(t1<t) break;} 
      /* Find the closest sample time after this one */ 
      for(i2=i; i2<tac->sampleNr; i2++) {t2=tac->x[i2]; if(t2>t) break;} 
      /* Mean sample distance */
      f=0.0;
      if(t1<t) f+=t-t1; else f+=t2-t;
      if(t2>t) f+=t2-t; else f+=t-t1;
      f*=0.5; if(f<=0.0) f=1.0;
      tac->w[i]=f; sumw+=f;
    }

  }

  /* scale weights so that sum of weights equals sample number */
  sumw/=(double)tac->sampleNr;
  for(int i=0; i<tac->sampleNr; i++) tac->w[i]/=sumw;

  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
/** Get the number of samples in TAC that have weight > 0.
    Missing (NaN) sample values are included as long as weight is not missing.
    If weights are not set, then nr of all samples is returned.
   @return The number of samples with weight above zero.
   @sa tacWByFreq, aicSS, parFreeNr
*/
unsigned int tacWSampleNr(
  /** Pointer to the TAC struct. */
  TAC *tac
) {
  if(tac==NULL || tac->tacNr<1 || tac->sampleNr<1) return((unsigned int)0);
  if(tac->weighting==WEIGHTING_OFF) return((unsigned int)tac->sampleNr);
  unsigned int n=0;
  for(int i=0; i<tac->sampleNr; i++) if(tac->w[i]>0.0) n++;
  if(n==0 && tac->weighting==WEIGHTING_UNKNOWN) n=(unsigned int)tac->sampleNr;
  return(n);
}
/*****************************************************************************/

/*****************************************************************************/
