/** @file tpctac.h
 *  @brief Header file for library libtpctac.
 *  @details Time-activity curve (TAC) processing and file i/o.
 *  @author Vesa Oikonen
 *  @copyright (c) Turku PET Centre
 */
#ifndef _TPCTAC_H_
#define _TPCTAC_H_
/*****************************************************************************/

/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/*****************************************************************************/
//#include "tpcextensions.h"
#include "tpccsv.h"
#include "tpcift.h"
#include "tpcisotope.h"
/*****************************************************************************/

/*****************************************************************************/
/** @public TAC file format codes; see tac_format */
typedef enum {
  TAC_FORMAT_UNKNOWN,     ///< Unknown format
  TAC_FORMAT_SIMPLE,      ///< x and y's with space delimiters
  TAC_FORMAT_DFT,         ///< Data format of Turku PET Centre
  TAC_FORMAT_IFT,         ///< Interfile-type data (supported for writing)
  TAC_FORMAT_NCI,         ///< Old Turku PET Centre format
  TAC_FORMAT_PMOD,        ///< PMOD TAC format
  TAC_FORMAT_CSV_INT,     ///< International CSV
  TAC_FORMAT_CSV_UK,      ///< UK CSV
  TAC_FORMAT_TSV_INT,     ///< International TSV (comma as decimal separator)
  TAC_FORMAT_TSV_UK,      ///< UK TSV (point as decimal separator)
  TAC_FORMAT_CPT,         ///< CPT format
  TAC_FORMAT_IDWC,        ///< Hammersmith tissue file format
  TAC_FORMAT_IF,          ///< Hammersmith input file format
  TAC_FORMAT_XML,         ///< XML format (supported for writing)
  TAC_FORMAT_HTML,        ///< HTML table format (supported for writing)
  TAC_FORMAT_SIF,         ///< Scan information file
  TAC_FORMAT_XELERIS,     ///< Xeleris format (reading supported) 
  TAC_FORMAT_INVEON,      ///< Inveon format (reading supported) 
  TAC_FORMAT_AMIDE,       ///< Amide format (reading supported)
  TAC_FORMAT_CARIMAS_TXT, ///< Carimas txt format (reading supported) 
  TAC_FORMAT_QVIEW,       ///< QView CSV TAC format (reading supported)
  TAC_FORMAT_MAT,         ///< Matlab matrix TAC format (reading supported)
  TAC_FORMAT_HRRT_HC,     ///< HRRT head curve format (reading supported)
  TAC_FORMAT_HRPLUS_HC,   ///< HR+ head curve format (reading supported)
  TAC_FORMAT_ABSS_SCANDITRONICS, ///< Scanditronics ABSS data; reading supported
  TAC_FORMAT_ABSS_GEMS,   ///< GEMS ABSS data; reading supported
  TAC_FORMAT_ABSS_ALLOGG_OLD, ///< ALLOGG ABSS data (old format); reading supported
  TAC_FORMAT_ABSS_ALLOGG, ///< ALLOGG ABSS data; reading supported
  TAC_FORMAT_BINARY,      ///< Binary format (currently not supported)
  TAC_FORMAT_LAST
} tacformat;
/*****************************************************************************/

/*****************************************************************************/
/** Concentrations, name, size etc for a single time-activity curve inside TAC struct.
   @sa TAC, PAR, PARR, PARN
 */
typedef struct TACC {
  /** Content type: TACID_UNKNOWN, TACID_REF, TACID_PLASMA, TACID_BLOOD,
      TACID_PARENT, TACID_METABOLITE, TACID_VOI, TACID_VOXEL */
  tactype   type;  
  /** VOI volume, ROI size, sample mass, or something like that. */
  double    size;
  /** Size unit: UNIT_UNKNOWN, UNIT_UL, ... */
  unit      sunit;
  /** Pointer to y (concentration) values. */
  double   *y;
  /** Temporary switch. */
  char      sw;
  /** Temporary switch. */
  char      sw2;
  /** Full name string, using '_' to separate any name tokens when necessary. */
  char      name[MAX_TACNAME_LEN+1];
} TACC;
/*****************************************************************************/
/** Definition for a set of TACs with common header information and x (sample time) values.
   @sa TACC, MTAC, PAR, CSV, IFT, tacInit, tacFree
 */
typedef struct TAC {
  /** Nr of samples (frames) in each TAC; note that more memory may be allocated than this. */
  int           sampleNr;
  /** Nr of TACCs (VOIs). */
  int           tacNr;
  /** File format code: TAC_FORMAT_* */  
  tacformat     format;
  /** Sample time: are sample start and end times available (1), or
      just the middle time point (0) */    
  int           isframe;
  /** Sample time (or middle frame time); units specified with tunit */
  double       *x;
  /** Sampling (frame) start time; units specified with tunit. */
  double       *x1;
  /** Sampling (frame) end time; units specified with tunit. */
  double       *x2;
  /** Unit of y (concentration) values: UNIT_* 
     @sa unit, tunit
    */
  int           cunit;
  /** Unit of x (sample time) values: UNIT_*
     @sa unit, cunit
   */
  int           tunit;
  /** Pointer to weight factors */
  double       *w;
  /** Variable indicating whether weights are present
      WEIGHTING_UNKNOWN, WEIGHTING_ON, WEIGHTING_OFF */
  weights       weighting;
  /** Pointers to y (concentration) data */
  TACC         *c;
  /** Internal variable: Number of allocated TACCs (VOIs) */
  int           _tacNr;
  /** Internal variable: Number of allocated samples */
  int           _sampleNr;
  /** Internal variable: main pointer to all allocated memory for the
      arrays x[], x1[], x2[], w[], and all y[]'s.
      These data arrays just point to this array.
      Allocated memory size is (4+_tacNr)*_sampleNr. */
  double       *_data;
  /** @brief Optional (but often useful) header information.
      @details Reserved keys:
       - studynr and study_number
       - timeunit and time_unit
       - unit and calibration_unit
       - voiname[index]
       - sizes
       - radiopharmaceutical
       - isotope
       - decay_correction and decay correction
       - injection_time and injection time
       - scan_start_time, scan start time, scan_start, scan start, and scantime
      @sa iftFree, iftFindKey, iftPut
   */
  IFT           h;
} TAC;
/*****************************************************************************/

/*****************************************************************************/
/** Definition for a list of TAC data structures. 
   @sa TAC, TACC, mtacInit, mtacFree, mtacAllocate, mtacAddTAC
*/
typedef struct MTAC {
  /** List of TAC data structures. */
  TAC *tac;
  /** Number of TAC structures in the list. */
  int nr;
  /** Allocated TAC list size, _nr >= nr. */
  int _nr;
} MTAC;
/*****************************************************************************/

/*****************************************************************************/
/* tac */
void tacInit(TAC *tac);
void taccInit(TACC *tacc);
void tacFree(TAC *tac);
void tacFreeExceptHeader(TAC *tac);
void taccFree(TACC *tacc);
int tacAllocate(TAC *tac, int sampleNr, int tacNr);
int tacAllocateMore(TAC *tac, int tacNr);
int tacCopyTacc(TACC *d1, TACC *d2, int sampleNr);
int tacCopyTaccdata(TACC *d1, TACC *d2, int sampleNr);
int tacCopyTacchdr(TACC *d1, TACC *d2);
int tacCopyHdr(TAC *tac1, TAC *tac2);
int tacIsSize(TAC *d);
int tacDuplicate(TAC *tac1, TAC *tac2);
int tacAllocateMoreSamples(TAC *tac, int addNr);
/*****************************************************************************/

/*****************************************************************************/
/* mtac */
void mtacInit(MTAC *mtac);
void mtacFree(MTAC *mtac);
int mtacAllocate(MTAC *mtac, int nr);
int mtacAllocateMore(MTAC *mtac, int nr);
int mtacAddTAC(MTAC *mtac, TAC *tac);
/*****************************************************************************/

/*****************************************************************************/
/* tacio */
char *tacFormattxt(tacformat c);
int tacFormatIdentify(const char *s);
char *tacDefaultExtension(tacformat c);
int tacFormatDetermine(const char *fname, TPCSTATUS *status);
int tacFormatWriteSupported(tacformat format);
int tacWrite(TAC *tac, FILE *fp, tacformat format, int extra, TPCSTATUS *status);
int tacRead(TAC *d, const char *fname, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* tacift */
int tacGetHeaderStudynr(IFT *h, char *s, TPCSTATUS *status);
int tacSetHeaderStudynr(IFT *h, const char *s);
int tacGetHeaderUnit(TAC *tac, TPCSTATUS *status);
int tacSetHeaderUnit(IFT *h, int u);
int tacGetHeaderTimeunit(TAC *tac, TPCSTATUS *status);
int tacSetHeaderTimeunit(IFT *h, int u);
int tacGetHeaderIsotope(IFT *h, char *s, TPCSTATUS *status);
int tacSetHeaderIsotope(IFT *h, const char *s);
int tacGetHeaderScanstarttime(IFT *h, char *s, TPCSTATUS *status);
int tacSetHeaderScanstarttime(IFT *h, const char *s);
decaycorrection tacGetHeaderDecayCorrection(IFT *h);
int tacSetHeaderDecayCorrection(IFT *h, decaycorrection dc);
/*****************************************************************************/

/*****************************************************************************/
/* simpleio */
int tacWriteSimple(TAC *tac, FILE *fp, int extra, TPCSTATUS *status);
int tacReadSimple(TAC *tac, CSV *csv, IFT *hdr, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* dftio */
int tacNameSplit(
  const char *rname, char *name1, char *name2, char *name3, int max_name_len
);
int tacWriteDFT(TAC *tac, FILE *fp, int extra, TPCSTATUS *status);
int tacReadDFT(TAC *tac, CSV *csv, IFT *hdr, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* sifio */
int tacWriteSIF(TAC *tac, FILE *fp, int extra, TPCSTATUS *status);
int tacReadSIF(TAC *tac, CSV *csv, IFT *hdr, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* pmodio */
int tacWritePMOD(TAC *tac, FILE *fp, int extra, TPCSTATUS *status);
int tacReadPMOD(TAC *tac, CSV *csv, IFT *hdr, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* qviewio */
int tacReadQView(TAC *tac, CSV *csv, const int grouponly, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* matio */
int tacReadMat(TAC *tac, CSV *csv, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* taccsv */
int tacWriteCSV(TAC *tac, FILE *fp, int extra, tacformat format, TPCSTATUS *status);
int tacReadCSV(TAC *tac, CSV *csv, IFT *hdr, TPCSTATUS *status);
int tacReadInveonCSV(TAC *tac, CSV *csv, TPCSTATUS *status);
int tacReadCarimasTxt(TAC *tac, CSV *csv, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* hcio */
int tacReadHRRTHC(TAC *tac, CSV *csv, TPCSTATUS *status);
int tacReadHRPLUSHC(TAC *tac, CSV *csv, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* tacnan */
int tacXNaNs(TAC *tac);
int tacYNaNs(TAC *tac, const int i);
int tacNaNs(TAC *tac);
int tacNotNaNs(TAC *tac, const int i);
int tacFixNaNs(TAC *tac);
int tacNthSample(TAC *tac, const int sn, const int i);
/*****************************************************************************/

/*****************************************************************************/
/* tacselect */
int tacSelectTACs(TAC *d, const char *region_name, int reset,
  TPCSTATUS *status
);
int tacSelectedTACs(TAC *d);
int tacSelectBestReference(TAC *d);
int tacNameMatch(const char *tacname, const char *test_str, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* tacorder */
int tacVerifyTimeOrder(TAC *d, TPCSTATUS *status);
int tacSortByTime(TAC *d, TPCSTATUS *status);
int tacSortByConc(TAC *d, const int i, TPCSTATUS *status);
int tacSortByName(TAC *d, TPCSTATUS *status);
int tacSortByAUC(TAC *d, TPCSTATUS *status);
int tacMoveTACC(TAC *d, int from, int to);
int tacSwapTACCs(TAC *d, int i1, int i2);
int tacDeleteTACC(TAC *d, int i);
int tacMultipleSamples(TAC *d1, const int fixMode, TAC *d2, const int verbose);
/*****************************************************************************/

/*****************************************************************************/
/* taccomp */
int tacCompareUnit(TAC *d1, TAC *d2, TPCSTATUS *status);
int tacCompareNames(TAC *d1, TAC *d2, const int i, TPCSTATUS *status);
int tacCompareConc(
  TAC *d1, TAC *d2, const int i, const double test_abs, const double test_rel,
  TPCSTATUS *status
);
int tacCompareTimes(
  TAC *d1, TAC *d2, const double test_abs, const double test_rel,
  TPCSTATUS *status
);
int tacCompareWeights(
  TAC *d1, TAC *d2, const double test_abs, const double test_rel,
  TPCSTATUS *status
);
/*****************************************************************************/

/*****************************************************************************/
/* tacunits */
int tacXUnitConvert(TAC *d, const int u, TPCSTATUS *status);
int tacYUnitConvert(TAC *d, const int u, TPCSTATUS *status);
int tacYUnitMass2Volume(TAC *tac, const double density, TPCSTATUS *status);
int tacYUnitVolume2Mass(TAC *tac, const double density, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* tacname */
int tacIndividualNames(TAC *tac);
void tacEnsureNames(TAC *tac);
/*****************************************************************************/

/*****************************************************************************/
/* tacx */
int tacXCopy(TAC *tac1, TAC *tac2, int i1, int i2);
int tacCorrectFrameOverlap(TAC *d, TPCSTATUS *status);
int tacXRange(TAC *d, double *xmin, double *xmax);
int tacSampleXRange(TAC *d, double *xmin, double *xmax);
int tacMinX(TAC *d);
int tacIsX(TAC *d);
int tacXMatch(TAC *d1, TAC *d2, const int verbose);
int tacAddZeroSample(TAC *d, TPCSTATUS *status);
int tacDeleteSample(TAC *d, int i);
int tacDeleteMissingSamples(TAC *d);
int tacExtractRange(TAC *d1, TAC *d2, double startT, double endT);
int tacExtractSamples(TAC *d1, TAC *d2, int si, int ei);
int tacCheckX1X2X(TAC *d);
int tacSetX(TAC *d, TPCSTATUS *status);
int tacGetSampleInterval(TAC *d, double ilimit, double *minfdur, double *maxfdur);
/*****************************************************************************/

/*****************************************************************************/
/* tacy */
int tacYRange(
  TAC *d, int i, double *ymin, double *ymax,
  int *smin, int *smax, int *imin, int *imax
);
int tacYRangeInXRange(
  TAC *d, int i, const double xmin, const double xmax, 
  double *ymin, double *ymax, int *smin, int *smax, int *imin, int *imax
);
/*****************************************************************************/

/*****************************************************************************/
/* tacw */
int tacWCopy(TAC *tac1, TAC *tac2, int i1, int i2);
int tacWMove(TAC *tac, int ow, TPCSTATUS *status);
int tacWByFreq(TAC *tac, TPCSTATUS *status);
unsigned int tacWSampleNr(TAC *tac);
/*****************************************************************************/

/*****************************************************************************/
/* tacdc */
int tacGetIsotope(TAC *tac);
void tacSetIsotope(TAC *tac, int isotope);
int tacDecayCorrection(TAC *tac, int isotope, int mode, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* tacabss */
int tacReadAllogg(TAC *tac, IFT *hdr, TPCSTATUS *status);
int tacReadOldAllogg(TAC *tac, IFT *hdr, TPCSTATUS *status);
int tacReadScanditronics(TAC *tac, IFT *hdr, TPCSTATUS *status);
int tacReadGEMS(TAC *tac, IFT *hdr, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* tacxmlio */
int tacWriteXML(TAC *tac, FILE *fp, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
#endif /* TPCTAC */
